import SWAL from "sweetalert2";
import * as THREE from "three";

window.window.document.addEventListener("DOMContentLoaded", (event: Event) => new Asynchronous_programming('img/Bus.jpg', 20, 5) /* Each 4 sec. within 20 sec. */);

class Asynchronous_programming {
    private readonly _image: HTMLImageElement = new Image();

    private _time_left; // In sec...

    private _animation_id: number = 0; // '0' means that animation hasn't yet started...
    private _interval_id!: number;

    private readonly _camera = new THREE.PerspectiveCamera;
    private readonly _renderer = new THREE.WebGLRenderer;
    private readonly _scene = new THREE.Scene();

    constructor(private readonly _image_URL: string, private readonly _count_down = 20, private readonly _steps = 4) {
        this._time_left = this._count_down; // In sec.

        window.document.addEventListener("Image_is_ready", this._create_3D_object, false);
        window.document.addEventListener("3D_object_is_ready", () => { // 'custom_event' is ignored
            this._interval_id = window.setInterval(this._time_out, this._count_down / this._steps * 1000);
        }, false);
        window.document.addEventListener("3D_object_is_ready", (custom_event: Event) => {
            if ((custom_event as CustomEvent).detail.name !== this._image_URL)
                throw ("Abnormal situation...");
            this._animation_id = window.requestAnimationFrame(this._animate_3D_object);
        }, false);

        this._image.onload = () => window.document.dispatchEvent(new Event("Image_is_ready"));
        this._image.src = this._image_URL;

        this._camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 0.1, 5000);
        this._camera.position.x += 100;
        this._camera.position.y += 100;
        this._camera.position.z = 3000;

        this._renderer = new THREE.WebGLRenderer({/*alpha: true,*/ antialias: true});
        this._renderer.setPixelRatio((window.devicePixelRatio) ? window.devicePixelRatio : 1); // If exists then OK else '1'

        this._renderer.setClearColor(new THREE.Color("rgb(211,211,211)"), 0.5); // Silver
        this._renderer.setSize(window.innerWidth, window.innerHeight);
        window.document.body.appendChild(this._renderer.domElement);
    }

    private _create_3D_object: () => void = () => {
// Test:
        window.console.assert(this._image && this._image instanceof Image && this._image.complete, 'Contract violation: \'this._image\'');
// Versus defensive programming:
        if (!(this._image && this._image instanceof Image && this._image.complete))
            throw ('Contract violation: \'this._image\'');

        const canvas = window.document.createElement('canvas');
        canvas.width = this._image.width;
        canvas.height = this._image.height;
        canvas.getContext('2d')!.drawImage(this._image, 0, 0, this._image.width, this._image.height);

        const texture = new THREE.CanvasTexture(canvas);
        texture.minFilter = THREE.LinearFilter;

        const mesh = new THREE.Mesh(new THREE.SphereGeometry(window.innerWidth / 2, 10, 10, 0, Math.PI), new THREE.MeshBasicMaterial({
            map: texture,
            opacity: 1.,
            side: THREE.DoubleSide,
            transparent: true
        }));
        mesh.name = this._image_URL;
        this._scene.add(mesh);

        window.document.dispatchEvent(new CustomEvent("3D_object_is_ready", {'detail': {name: this._image_URL}}));
    }

    private _animate_3D_object = (timestamp?: number) => {
        if (timestamp !== undefined) { /* first call: 'timestamp === undefined' */
            // One may handle the elapsed time from the first call to '_animate_3D_object'
        }
        (this._scene.getObjectByName(this._image_URL) as THREE.Mesh).rotation.y -= 0.01;
        ((this._scene.getObjectByName(this._image_URL) as THREE.Mesh).material as THREE.MeshStandardMaterial).opacity -= 0.0008;
        this._renderer.render(this._scene, this._camera);

        this._animation_id = window.requestAnimationFrame(this._animate_3D_object);
    }

    private _time_out = () => {
        if ((this._time_left -= this._count_down / this._steps) > 0)
            SWAL.fire({
                showConfirmButton: false,
                title: 'You can miss the bus...',
                text: this._time_left + ' sec.',
                timer: 1000
            });
        else {
            ((this._scene.getObjectByName(this._image_URL) as THREE.Mesh).material as THREE.Material).opacity = 1;
            ((this._scene.getObjectByName(this._image_URL) as THREE.Mesh).material as THREE.MeshStandardMaterial).color.setHex(0xFF0000);
            this._renderer.render(this._scene, this._camera);

            window.cancelAnimationFrame(this._animation_id);
            window.clearInterval(this._interval_id);

            SWAL.fire({
                showConfirmButton: false,
                title: 'You miss it!',
                text: this._time_left + ' sec.',
                timer: 1000
            });
        }
    }
}






